/*
 * rpc_xdr.c - 
 *
 *	$Id: rpc_xdr.c,v 1.11 1993/05/05 15:37:27 idl Exp $
 */

/*
  Copyright (c) 1992-1993, Research Systems Inc.  All rights reserved.
  This software includes information which is proprietary to and a
  trade secret of Research Systems, Inc.  It is not to be disclosed
  to anyone outside of this organization. Reproduction by any means
  whatsoever is  prohibited without express written permission.
  */

#if !defined(MAC)

#ifndef IDL_CLIENT
#include "basic.h"
#endif
#include <stdio.h>
#include <ctype.h>
#include <rpc/rpc.h>
#include "export.h"
#include "rpc_idl.h"
#ifdef IDL_CLIENT
#include "xdr_supp.h"
#else
#include "xdr_supp.x"
#endif

/************************************************************************/


#ifndef IDL_CLIENT
#define		MAKE_TEMP		make_temp_array
#define		STR_ENSURE		str_ensure_length
#endif

#ifdef IDL_CLIENT

#define		MAKE_TEMP		make_temp
#define		STR_ENSURE		str_ensure
#define gettmp()                (VPTR)malloc(sizeof(VARIABLE))

/*
 *  make_temp_array
 *      This is the client side version of the make_temp_array
 *      We just malloc the amount desired and point var->value.arr->data.
 *      We assume that var is already valid except for the array data.
 *
 *      We also assume that user of this function will finish with
 *      the variable before calling make_temp_array again.
 *      Of course, this function is only locally accessible so that
 *      can never happen ;-)
 */

static  char*   make_temp( type, n_dim, dim, init, var )
    int         type;
    int         n_dim;
    LONG        dim[];
    int         init;	/* IGNORED */
    VPTR        *var;
{
    UCHAR               * data;
    ARRAY               * arr;
    ULONG       	arr_size;
    ULONG	        elem_size;
    int                 i;
    VPTR                newvar;

    switch( type ) {

      case TYP_BYTE:    arr_size        = sizeof(char);         break;
      case TYP_INT:     arr_size        = sizeof(short);        break;
      case TYP_LONG:    arr_size        = sizeof(LONG);         break;
      case TYP_FLOAT:   arr_size        = sizeof(float);        break;
      case TYP_DOUBLE:  arr_size        = sizeof(double);       break;
      case TYP_COMPLEX: arr_size        = sizeof(COMPLEX);      break;
      case TYP_STRING:  arr_size        = sizeof(STRING);       break;

      default:		/*      Error   */
	return NULL;
    }

    elem_size   = arr_size;		/* save for later use           */

    for( i = 0; i < n_dim; ++i ) {
	if( dim[ i ] <= 0 )
	    return NULL;
	arr_size        *= dim[i];
    }

    if( (arr = (ARRAY*)malloc( sizeof(ARRAY) )) == NULL ) {
	return NULL;
    }

    if( (data = (UCHAR*)malloc( arr_size )) == NULL ) {
	free( arr );
	return NULL;
    }

    /*  Fill everything in */

    *var                =
      newvar            = (VPTR)malloc(sizeof(VARIABLE));
    if( newvar == NULL ) {
	free(arr);
	free(data);
	return NULL;
    }
    newvar->type        = type;
    newvar->flags       = V_ARR | V_DYNAMIC;
    newvar->value.arr   = arr;

    arr->elt_len        = elem_size;
    arr->arr_len        = arr_size;
    arr->n_elts         = arr_size / elem_size;
    arr->data           = data;
    arr->n_dim          = n_dim;
    arr->flags          = 0;
    arr->file_unit      = 0;
    bcopy( (char *)dim, (char *)arr->dim, n_dim * sizeof(LONG) );

    if( type == TYP_STRING ) {
	bzero( (char *)data, arr_size );
    }
    return( (char*)data );
}

static void str_ensure( p, length )
     STRING     *p;
     int        length;
{
    int reallocate = ((int) p->slen < length);

    /* Delete existing string, if need more room, or if result is null */

    if ((!length && p->slen) || (length && reallocate)) {
	if( p->slen )
	    free( p->s );
	p->s    = NULL;
    }

    if (reallocate) {	/* The existing string is not long enough */
	if( (p->s = (char*)malloc(length+1)) == NULL ) {
	    p->slen     = 0;
	    return;
	}
	p->stype = 1;	/* Type = dynamic */
    }

    p->slen = length;			/* Set length */
}



#if defined(mips) || defined(AUX)
public bool_t xdr_idl_mips_double(XDR *xdrs, double *p)
/*
 * DecStation 3100 and other mips machines trash XDR doubles. This routine uses
 * xdr_long() to output the double properly. This works because the
 * mips uses ieee math, which happens to be the XDR standard.
*/
{
#ifdef mips_ultrix                     /* Little Endian */
  return(XDR_LONG(xdrs, ((LONG *) p) + 1) && XDR_LONG(xdrs, (LONG *) p));
#else                                  /* Big endian */
  return(XDR_LONG(xdrs, (LONG *) p) && XDR_LONG(xdrs, ((LONG *) p + 1)));
#endif
}
#endif



#if defined(vax) && defined(ultrix)
public bool_t xdr_idl_vax_ultrix_float(XDR *xdrs, float *p)
/*
 * Ultrix vax XDR returns a 0x00010000 bit pattern when the actual
 * value is zero. This routine calls xdr_float(), checks for this
 * value, and patches in a zero if necessary.
*/
{
  bool_t status;

  status = xdr_float(xdrs, p);
  if ((xdrs->x_op == XDR_DECODE) && status &&
      (0x00000100 == *((ULONG *) p))) *p = 0.0;
  return(status);
}
#endif


/*
 *  free_idl_var
 */
public  void    free_idl_var( varinfo_t* v )
{
    ULONG	       n;
    ARRAY *            arr;
    STRING *           data;

    /*  Do nothing if there is no VARIABLE structure or the user has    */
    /*  indicated that the variable is statically allocated             */

    if( v->Variable == NULL || v->Length != 0)
	    return;
    
    /*  Is variable an array variable?                                  */
    
    if( VarIsArray(v) ) {

	arr     = v->Variable->value.arr;

	/*      If we have an array of strings, each string contains    */
	/*      allocated data                                          */

	if( v->Variable->type == TYP_STRING ) {
	    data        = (STRING *)arr->data;
	    n           = arr->n_elts;
	    while( n-- ) {
		if( data->slen )
		    free( data->s );
		++data;
	    }
	}

	free( arr->data );			/* Free raw data        */
	free( v->Variable->value.arr );		/* Free array struct    */
    }

    /*  Is variable a STRING and is the string non-NULL?                */
    
    else if( GetVarType(v) == TYP_STRING && v->Variable->value.str.slen ) {

	/*      The only dynamically allocated scalars are strings      */

	free( v->Variable->value.str.s );	/* Free String pointer  */
    }

    free( v->Variable );	/* Free variable struct                 */
    v->Variable = 0;
}


#endif  /* IDL_CLIENT */


/************************************************************************/


/*
 *  xdr_idl_getinfo
 */
public bool_t xdr_idl_getinfo( xdrsp, getinfo )
    XDR         * xdrsp;
    getinfo_t   * getinfo;
{
    u_int       max_elem        = MAXIDLEN + 1;
    char *      p               = getinfo->Name;

    return(xdr_bytes( xdrsp, &p, &max_elem, max_elem ) &&
	   xdr_short( xdrsp, &getinfo->TypeCode ));
}


/*
 *  xdr_idl_complex
 *      This routine is copied from xdr_supp.c
 */
public bool_t xdr_idl_rpc_complex( xdrsp, p )
    XDR         * xdrsp;
    COMPLEX     * p;
{
    return(XDR_FLOAT(xdrsp, (XDR_FLOAT_TYPE *) &(p->r))
	   && XDR_FLOAT(xdrsp, (XDR_FLOAT_TYPE *) &(p->i)));
}



/*
 *  xdr_idl_string
 */
public bool_t xdr_idl_rpc_string( xdrsp, p )
    XDR         * xdrsp;
    STRING      * p;
{
    short length = p->slen;

    /* R/W the length */
    if (!xdr_short(xdrsp, (short *) &length)) return(FALSE);

    /* If reading, make sure the string is long enough. This step also
       sets the length in the string descriptor. */

    if (xdrsp->x_op == XDR_DECODE) {
	STR_ENSURE(p, length);
	if( length && p->s == NULL ) return( 0 );
    }

    /* R/W the string --- but only if it's non-null. */

    return(length ? xdr_string(xdrsp, &p->s, length) : TRUE);
}

/*
 *  xdr_idl_rpc_array
 *      transport an array.
 */
static int      xdr_idl_rpc_array(xdrsp, vinfo, t_var)
    XDR         *xdrsp;
    varinfo_t   *vinfo;
    VPTR        t_var;
{
    VPTR        old;
    int         status;
    ARRAY       *arr;
    char        *dim_addr;
    u_int       max_dim;

    xdrproc_t   xdr_proc;
    char        *data_addr;
    u_int       n_elts;

    /*  temporary storage to read stuff in so that we can allocate      */
    /*  it properly (which only happens *AFTER* we have read in         */
    /*  the information                                                 */

    ARRAY       t_arr;


    if( xdrsp->x_op != XDR_DECODE ) {
	t_arr   = *t_var->value.arr;
    }

    max_dim     = MAX_ARRAY_DIM;
    dim_addr    = (char *)t_arr.dim;

    status      = XDR_LONG(xdrsp, LONGA(t_arr.elt_len)) &&
	   XDR_LONG(xdrsp, LONGA(t_arr.arr_len)) &&
	   XDR_LONG(xdrsp, LONGA(t_arr.n_elts)) &&
	   xdr_u_char(xdrsp, UCHARA(t_arr.n_dim)) &&
	   xdr_u_char(xdrsp, UCHARA(t_arr.flags)) &&
	   xdr_short(xdrsp, SHORTA(t_arr.file_unit)) &&
	   xdr_array(xdrsp, &dim_addr, &max_dim, (u_int)max_dim,
		     sizeof(LONG), XDR_LONG);

    if( status == 0 )
	return 0;

    /*  Now get the raw data                                    */

    switch( t_var->type ) {

      case TYP_BYTE:    xdr_proc        = xdr_u_char;           break;
      case TYP_INT:     xdr_proc        = xdr_short;            break;
      case TYP_LONG:    xdr_proc        = XDR_LONG;             break;
      case TYP_FLOAT:   xdr_proc        = XDR_FLOAT;            break;
      case TYP_DOUBLE:  xdr_proc        = XDR_DOUBLE;           break;
      case TYP_COMPLEX: xdr_proc        = xdr_idl_rpc_complex;  break;
      case TYP_STRING:  xdr_proc        = xdr_idl_rpc_string;   break;
      default:
	/*      error                                   */
	return 0;
    }

    /*  If we are receiving data, build a temp array to put the         */
    /*  data in. Otherwise, just use what the user gave us              */

    if (xdrsp->x_op == XDR_DECODE ) {

	/*      Only allocate vinfo->Variable if we need to.            */
	/*      If the existing array is of the same type, with the     */
	/*      same number of elements, same n_dim and dims, we don't  */
	/*      reallocate.                                             */

#ifdef IDL_CLIENT

	if((old = vinfo->Variable) != NULL) {

	    /*  was previous variable an array?                         */
	    if( old->flags & V_ARR ) {

		/* Was previous entry statically allocated?             */
		if( vinfo->Length ) {
		    /*  Is static location too small to fit array?      */

		    if( vinfo->Length < t_arr.arr_len )
			return 0;

		    /* Fill in static array                             */

		    old->type           = t_var->type;
		    arr                 = old->value.arr;
		    arr->elt_len        = t_arr.elt_len;
		    arr->arr_len        = t_arr.arr_len;
		    arr->n_elts         = t_arr.n_elts;
		    arr->n_dim          = t_arr.n_dim;
		    bcopy((char *) t_arr.dim,
			  (char *) arr->dim,
			  sizeof(LONG)*t_arr.n_dim);

		    /*  Get raw data address                            */
		    data_addr           = (char *) arr->data;
		}

		/*      Was previous dynamically allocated array the    */
		/*      perfect size to fit the current data?           */
		else if(t_var->type == old->type &&
		    t_arr.n_dim == old->value.arr->n_dim &&
		    !bcmp((char *)t_arr.dim, (char *)old->value.arr->dim,
			  sizeof(LONG)*t_arr.n_dim)) {
		    data_addr   = (char *) old->value.arr->data;
		}
		else {
		    free_idl_var(vinfo);
		    data_addr   = MAKE_TEMP((int)t_var->type,
					    t_arr.n_dim,
					    t_arr.dim,
					    BASICARR_INI_NOP,
					    &vinfo->Variable );
		}
	    }
	    else {
		free_idl_var(vinfo);
		data_addr       = MAKE_TEMP((int)t_var->type,
					    t_arr.n_dim,
					    t_arr.dim,
					    BASICARR_INI_NOP,
					    &vinfo->Variable );
	    }
	}
	else

#endif
	    /*  If there is no adequate space available from a          */
	    /*  previous allocation (always the case if we are the      */
	    /*  server) then dynamically allocate more space            */

	    data_addr   = MAKE_TEMP((int)t_var->type,
				    t_arr.n_dim,
				    t_arr.dim,
				    BASICARR_INI_NOP,
				    &vinfo->Variable );
    }
    else {
	data_addr       = CHAR(t_arr.data);
    }

    /*  Check for allocation failure                            */

    if( vinfo->Variable == NULL || data_addr == NULL )
	return 0;

    n_elts      = t_arr.n_elts;

    /*  XDR the array elements                                  */

    return( xdr_array(xdrsp, &data_addr, &n_elts, n_elts,
				t_arr.elt_len, xdr_proc));
}


/*
 *  xdr_vinfo
 *
 *      Assume a varinfo_t looks like:
 *
 *      struct {
 *         char         * Name;
 *         VARIABLE     Variable;
 *      }
 */
xdr_vinfo( xdrsp, vinfo)
    XDR         * xdrsp;
    varinfo_t   * vinfo;
{
    int         status;
    unsigned int                max_elem;
    VPTR        v;
    short       istmp;
    char        *p;

    /*  temporary storage to read stuff in so that we can allocate      */
    /*  it properly (which only happens *AFTER* we have read in         */
    /*  the information                                                 */

    VARIABLE    t_var;

    /*          Get the variable name                                   */

    max_elem    = MAXIDLEN+1;
    p           = vinfo->Name;
    if( !xdr_bytes( xdrsp, &p, &max_elem, max_elem) )
	return 0;
    
    /*  If we are *not* receiving information we need to use the        */
    /*  info provided by the user                                       */

    if( xdrsp->x_op != XDR_DECODE ) {
	t_var   = *vinfo->Variable;
    }
    else
	bzero((char *)&t_var, sizeof(t_var));

    /*          Now start on the VARIABLE structure. type and flags     */

    if( !xdr_u_char( xdrsp, &t_var.type ) )
	return 0;
    if( !xdr_u_char( xdrsp, &t_var.flags ) )
	return 0;

    if( t_var.flags & V_ARR ) {
	return xdr_idl_rpc_array( xdrsp, vinfo, &t_var );
    }

    /*  Don't clobber static old static arrays                          */

    if( xdrsp->x_op == XDR_DECODE && vinfo->Length != 0 )
	return 0;

    switch( t_var.type ) {

      case TYP_UNDEF:
	status = TRUE;
	break;

      case TYP_BYTE:
	status  = xdr_u_char( xdrsp, &t_var.value.c );
	break;

      case TYP_INT:
	status  = xdr_short( xdrsp, &t_var.value.i );
	break;

      case TYP_LONG:
	status  = XDR_LONG( xdrsp, &t_var.value.l );
	break;

      case TYP_FLOAT:
	status  = XDR_FLOAT( xdrsp, (XDR_FLOAT_TYPE *) &t_var.value.f );
	break;

      case TYP_DOUBLE:
	status  = XDR_DOUBLE( xdrsp, &t_var.value.d );
	break;

      case TYP_COMPLEX:
	status  = xdr_idl_rpc_complex( xdrsp, &t_var.value.cmp );
	break;

      case TYP_STRING:
	status  = xdr_idl_rpc_string( xdrsp, &t_var.value.str );
	break;

      default:
	return 0;		/*      error           */
    }

    if( status == 0 ) return 0;

    /*  Fill in the vinfo structure                                     */
    /*  We are always dealing with a scalar variable at this point      */
    /*  If the Variable is NULL, allocate it. If it isn't, free stuff   */
    /*  before clobbering it. In the server, v is always NULL           */

    if (xdrsp->x_op == XDR_DECODE) {

#ifdef IDL_CLIENT
	if((v = vinfo->Variable) != NULL)
	    free_idl_var(vinfo);
#endif
	vinfo->Variable = v = gettmp();

    }
    else {
	v               = vinfo->Variable;
    }

    if( v == NULL ) return 0;

    istmp       = v->flags & V_TEMP;
    *v          = t_var;
    v->flags    |= istmp;

    return 1;
}

#endif
